import numpy as np


class Evaluator:
    """Evaluator.

    Base class that evaluates individuals. Must be implemented
    to match Individual implementation.
    """

    def __init__(self):
        pass

    def eval(self, individual):
        """eval.

        Evaluate individual, returning a numerical loss value.

        individual: Individual to evaluate.
        """
        if individual.loss is not None:
            return individual.loss
        else:
            raise NotImplementedError()
            # write loss value to individual

    def eval_all(self, individuals):
        """eval_all.

        Evaluate individuals in a given iterable, return numpy array of loss values.

        individuals: iterable containing Individuals.
        """
        return np.fromiter([ self.eval(ind) for ind in individuals ], dtype=np.float64)



default_config = {
    
    # population sizes
    'parents': 1,
    'offspring': 10,
    
    # enable/disable mutation and crossover
    'do_mutate': True,
    'do_crossover': False,

    # set mutation/crossover parameters
    'crossover_prob': 0.5,
    'mutation_prob': 1.0,

    # additional arguments
    'crossover_kwargs': {},
    'mutation_kwargs': {},
    
    'do_timeout_heuristic': False,
    'timeout': 100,
    'mutation_increment': 1.01,
    'max_mutation_rate': 0.98,

    # selection
    'selection': 'plus', # options: ['plus', 'comma']

    # multiprocessing options
    'kernels': 1
}



import numpy as np

from .individual import Individual
from .evaluator import Evaluator
from .optimizer import Optimizer


N = 10 # size of vector


class RealVector(Individual):

    def __init__(self, vec):
        super().__init__()
        self.vec = vec

    def __repr__(self):
        return str(self.vec) + " (loss={})".format(self.loss)

    def copy(self):
        ind = self.__class__(self.vec.copy())
        ind.loss = self.loss
        return ind

    @classmethod
    def random(cls, size):
        return cls(np.random.normal(loc=0, scale=2, size=N))

    def mutate(self, mutation_rate=0.1, mutation_width=1):       
        if isinstance(mutation_width, int):
            # interpret as absolute number
            k = mutation_width
        elif isinstance(mutation_width, float):
            # interpret as probability per vector element
            k = max(1, np.random.binomial(self.vec.size, mutation_width))

        indices = np.random.choice(self.vec.size, size=k, replace=False)
        noise = np.random.normal(loc=0, scale=mutation_rate, size=k)
        self.vec[indices] += noise
        self.loss = None

    def cross(self, other, crossover_rate=0.5):
        # uniform crossover
        k = max(1, np.random.binomial(self.vec.size, crossover_rate))
        indices = np.random.choice(self.vec.size, size=k, replace=False)
        self.vec[indices] = other.vec[indices]
        self.loss = None


class RealVectorEvaluator(Evaluator):

    def __init__(self, target_vector):
        self.target_vec = target_vector

    def eval(self, ind):
        if ind.loss is not None:
            return ind.loss
        else:
            loss = np.linalg.norm(ind.vec - self.target_vec, 2)
            ind.loss = loss
            return loss


def print_lines(ls):
    for l in ls:
        print(l)

conf = {
    'parents': 5,
    'offspring': 10,
    'do_crossover': True,
    'mutation_prob': 0.9,
    'crossover_kwargs': {
        'crossover_rate': 0.5},
    'mutation_kwargs': {
        'mutation_rate': 1,
        'mutation_width': 0.1}
}

ev = RealVectorEvaluator(np.random.uniform(-20, 20, size=N))
print('target: {}'.format(ev.target_vec))
opt = Optimizer(ev, config=conf)
res = opt.run([RealVector.random(N) for _ in range(5)], generations=1000)
print_lines(res)



import numpy as np

from .config         import default_config
from multiprocessing import Pool
from tqdm            import trange


def coinflip(p=0.5):
    return np.random.binomial(1, p) == 1


class Optimizer:

    def __init__(self, evaluator, config=None):
        self.evaluator = evaluator

        # use defaults and overwrite with custom options
        self.config = default_config.copy()
        if config is not None:
            self.config.update(config)
            
        self.check_config()

    def check_config(self):
        """check_config.
        
        Check if the configuration is valid, and raise error if there are any
        inconsistencies or impossible combinations.
        """
        if not (1 <= self.config['parents'] <= self.config['offspring']):
            raise ValueError('`parents` and `offspring` must be positive integers, with `offspring` being greater or equal to `parents`')
        
        if (self.config['parents'] == 1) and self.config['do_crossover']:
            raise ValueError('crossover can only be performed with at least 2 `parents`')
        
        if self.config['selection'] not in ['plus', 'comma']:
            raise ValueError('unknown selection {}, must be \'plus\' or \'comma\''.format(self.config['selection']))

    def run(self, initial, generations=10):

        npar = self.config['parents']
        noff = self.config['offspring']

        # initialize population
        if len(initial) != npar:
            raise ValueError("given initial population size {} does not match parent population size {}".format(len(initial), npar))
        parents = [ ind.copy() for ind in initial ]
        
        # initially evaluate parents
        self.evaluator.eval_all(parents)
        
        # initially sort parents
        parents.sort(key=lambda ind: ind.loss)
        #parent_loss, parents = zip(*sorted(zip(parent_loss, parents)))
        
                
        def augment(args):
            ix, ind = args
            if self.config['do_crossover'] and coinflip(self.config['crossover_prob']):
                ind.cross(parents[(ix + np.random.randint(1, npar)) % npar], **self.config['crossover_kwargs'])
            if self.config['do_mutate'] and coinflip(self.config['mutation_prob']):
                ind.mutate(**self.config['mutation_kwargs'])
            return ind
        
        offspring = []
        for generation in trange(generations):
            
            # sample random parent indices
            parent_ixs = np.random.randint(0, npar, size=noff)
            
            for ix in parent_ixs:
                ind = parents[ix].copy()
                
                if self.config['do_crossover'] and coinflip(self.config['crossover_prob']):
                    ind.cross(parents[(ix + np.random.randint(1, npar)) % npar], **self.config['crossover_kwargs'])
                if self.config['do_mutate'] and coinflip(self.config['mutation_prob']):
                    ind.mutate(**self.config['mutation_kwargs'])

                offspring.append(ind)

            # evaluate offspring and sort by loss
            self.evaluator.eval_all(offspring)
            offspring.sort(key=lambda ind: ind.loss)
            #offspring_loss, offspring = zip(*sorted(zip(self.evaluator.eval_all(offspring), offspring)))

            # sort better offspring into parent population,
            # preferring offspring when loss is equal;
            # this uses the fact that parents and offspring are internally sorted
            pix, oix = 0, 0
            while (pix < npar) and (oix < noff):
                if parents[pix].loss >= offspring[oix].loss:
                    parents[pix] = offspring[oix].copy()
                    oix += 1
                pix += 1
                
            offspring.clear()
            
        return parents
